@@ -30,6 +30,9 @@
 
 
 /* Constants */
+#ifndef EAUTH
+# define EAUTH 80
+#endif
 
 #ifndef HOST_NAME_MAX
 # ifdef _POSIX_HOST_NAME_MAX
@@ -471,10 +474,6 @@ typedef uint16_t	in_port_t;
 #  error "Neither EAI_NODATA and EAI_NONAME are defined! :("
 # endif
 #endif
-
-#ifndef CLOCK_MONOTONIC
-#define CLOCK_MONOTONIC -1
-#endif
 /* end of chl */
 
 #ifndef HAVE_FPARSELN
@@ -507,8 +506,4 @@ typedef uint16_t	in_port_t;
 #define LOG_PERROR 0
 #endif
 
-#if !HAVE_DECL_SSL_OP_NO_CLIENT_RENEGOTIATION
-#define SSL_OP_NO_CLIENT_RENEGOTIATION 0
-#endif
-
 #endif /* _DEFINES_H */
