@@ -1,4 +1,4 @@
-/*	$OpenBSD: smtpctl.c,v 1.154 2017/07/27 18:48:30 sunil Exp $	*/
+/*	$OpenBSD: smtpctl.c,v 1.165 2019/07/23 08:11:10 gilles Exp $	*/
 
 /*
  * Copyright (c) 2013 Eric Faurot <eric@openbsd.org>
@@ -41,14 +41,21 @@
 #include <errno.h>
 #include <event.h>
 #include <fts.h>
+#include <grp.h>
 #include <imsg.h>
 #include <inttypes.h>
 #include <pwd.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <syslog.h>
 #include <time.h>
 #include <unistd.h>
+#if defined(HAVE_VIS_H) && !defined(BROKEN_STRNVIS)
+#include <vis.h>
+#else
+#include "bsd-vis.h"
+#endif
 #include <limits.h>
 
 #include "smtpd.h"
@@ -86,6 +93,8 @@ static int is_gzip_buffer(const char *);
 static FILE *offline_file(void);
 static void sendmail_compat(int, char **);
 
+extern int	spfwalk(int, struct parameter *);
+
 extern char	*__progname;
 int		 sendmail;
 struct smtpd	*env;
@@ -139,7 +148,7 @@ srv_connect(void)
 		return (0);
 	}
 
-	ibuf = xcalloc(1, sizeof(struct imsgbuf), "smtpctl:srv_connect");
+	ibuf = xcalloc(1, sizeof(struct imsgbuf));
 	imsg_init(ibuf, ctl_sock);
 
 	return (1);
@@ -261,6 +270,15 @@ srv_get_string(const char **s)
 	if (rlen == 0)
 		errx(1, "message too short");
 
+	rlen -= 1;
+	if (*rdata++ == '\0') {
+		*s = NULL;
+		return;
+	}
+
+	if (rlen == 0)
+		errx(1, "bogus string");
+
 	end = memchr(rdata, 0, rlen);
 	if (end == NULL)
 		errx(1, "unterminated string");
@@ -493,6 +511,29 @@ srv_show_cmd(int cmd, const void *data, size_t len)
 	} while (!done);
 }
 
+static void
+droppriv(void)
+{
+	struct passwd *pw;
+
+	if (geteuid())
+		return;
+
+	if ((pw = getpwnam(SMTPD_USER)) == NULL)
+		errx(1, "unknown user " SMTPD_USER);
+
+	if ((setgroups(1, &pw->pw_gid) ||
+	    setresgid(pw->pw_gid, pw->pw_gid, pw->pw_gid) ||
+	    setresuid(pw->pw_uid, pw->pw_uid, pw->pw_uid)))
+		err(1, "cannot drop privileges");
+}
+
+static int
+do_permission_denied(int argc, struct parameter *argv)
+{
+	errx(1, "need root privileges");
+}
+
 static int
 do_log_brief(int argc, struct parameter *argv)
 {
@@ -921,6 +962,8 @@ do_encrypt(int argc, struct parameter *argv)
 {
 	const char *p = NULL;
 
+	droppriv();
+
 	if (argv)
 		p = argv[0].u.u_str;
 	execl(PATH_ENCRYPT, "encrypt", p, (char *)NULL);
@@ -1005,92 +1048,81 @@ do_discover(int argc, struct parameter *argv)
 }
 
 static int
-do_uncorrupt(int argc, struct parameter *argv)
+do_spf_walk(int argc, struct parameter *argv)
 {
-	uint32_t msgid;
-	int	 ret;
+	droppriv();
 
-	if (ibuf == NULL && !srv_connect())
-		errx(1, "smtpd doesn't seem to be running");
-
-	msgid = argv[0].u.u_msgid;
-	srv_send(IMSG_CTL_UNCORRUPT_MSGID, &msgid, sizeof msgid);
-	srv_recv(IMSG_CTL_UNCORRUPT_MSGID);
-
-	if (rlen == 0) {
-		srv_end();
-		return (0);
-	} else {
-		srv_read(&ret, sizeof ret);
-		srv_end();
-	}
-
-	printf("command %s\n", ret ? "succeeded" : "failed");
-	return (0);
+	return spfwalk(argc, argv);
 }
 
+#define cmd_install_priv(s, f) \
+	cmd_install((s), privileged ? (f) : do_permission_denied)
+
 int
 main(int argc, char **argv)
 {
 	gid_t		 gid;
+	int		 privileged;
 	char		*argv_mailq[] = { "show", "queue", NULL };
 
 	__progname = ssh_get_progname(argv[0]);
 
 	sendmail_compat(argc, argv);
-	if (geteuid())
-		errx(1, "need root privileges");
+	privileged = geteuid() == 0;
 
 	gid = getgid();
 	if (setresgid(gid, gid, gid) == -1)
 		err(1, "setresgid");
 
-	cmd_install("discover <evpid>",		do_discover);
-	cmd_install("discover <msgid>",		do_discover);
+	/* Privileged commands */
+	cmd_install_priv("discover <evpid>",	do_discover);
+	cmd_install_priv("discover <msgid>",	do_discover);
+	cmd_install_priv("pause mta from <addr> for <str>", do_block_mta);
+	cmd_install_priv("resume mta from <addr> for <str>", do_unblock_mta);
+	cmd_install_priv("show mta paused",	do_show_mta_block);
+	cmd_install_priv("log brief",		do_log_brief);
+	cmd_install_priv("log verbose",		do_log_verbose);
+	cmd_install_priv("monitor",		do_monitor);
+	cmd_install_priv("pause envelope <evpid>", do_pause_envelope);
+	cmd_install_priv("pause envelope <msgid>", do_pause_envelope);
+	cmd_install_priv("pause envelope all",	do_pause_envelope);
+	cmd_install_priv("pause mda",		do_pause_mda);
+	cmd_install_priv("pause mta",		do_pause_mta);
+	cmd_install_priv("pause smtp",		do_pause_smtp);
+	cmd_install_priv("profile <str>",	do_profile);
+	cmd_install_priv("remove <evpid>",	do_remove);
+	cmd_install_priv("remove <msgid>",	do_remove);
+	cmd_install_priv("remove all",		do_remove);
+	cmd_install_priv("resume envelope <evpid>", do_resume_envelope);
+	cmd_install_priv("resume envelope <msgid>", do_resume_envelope);
+	cmd_install_priv("resume envelope all",	do_resume_envelope);
+	cmd_install_priv("resume mda",		do_resume_mda);
+	cmd_install_priv("resume mta",		do_resume_mta);
+	cmd_install_priv("resume route <routeid>", do_resume_route);
+	cmd_install_priv("resume smtp",		do_resume_smtp);
+	cmd_install_priv("schedule <msgid>",	do_schedule);
+	cmd_install_priv("schedule <evpid>",	do_schedule);
+	cmd_install_priv("schedule all",	do_schedule);
+	cmd_install_priv("show envelope <evpid>", do_show_envelope);
+	cmd_install_priv("show hoststats",	do_show_hoststats);
+	cmd_install_priv("show message <msgid>", do_show_message);
+	cmd_install_priv("show message <evpid>", do_show_message);
+	cmd_install_priv("show queue",		do_show_queue);
+	cmd_install_priv("show queue <msgid>",	do_show_queue);
+	cmd_install_priv("show hosts",		do_show_hosts);
+	cmd_install_priv("show relays",		do_show_relays);
+	cmd_install_priv("show routes",		do_show_routes);
+	cmd_install_priv("show stats",		do_show_stats);
+	cmd_install_priv("show status",		do_show_status);
+	cmd_install_priv("trace <str>",		do_trace);
+	cmd_install_priv("unprofile <str>",	do_unprofile);
+	cmd_install_priv("untrace <str>",	do_untrace);
+	cmd_install_priv("update table <str>",	do_update_table);
+
+	/* Unprivileged commands */
 	cmd_install("encrypt",			do_encrypt);
 	cmd_install("encrypt <str>",		do_encrypt);
-	cmd_install("pause mta from <addr> for <str>", do_block_mta);
-	cmd_install("resume mta from <addr> for <str>", do_unblock_mta);
-	cmd_install("show mta paused",		do_show_mta_block);
-	cmd_install("log brief",		do_log_brief);
-	cmd_install("log verbose",		do_log_verbose);
-	cmd_install("monitor",			do_monitor);
-	cmd_install("pause envelope <evpid>",	do_pause_envelope);
-	cmd_install("pause envelope <msgid>",	do_pause_envelope);
-	cmd_install("pause envelope all",	do_pause_envelope);
-	cmd_install("pause mda",		do_pause_mda);
-	cmd_install("pause mta",		do_pause_mta);
-	cmd_install("pause smtp",		do_pause_smtp);
-	cmd_install("profile <str>",		do_profile);
-	cmd_install("remove <evpid>",		do_remove);
-	cmd_install("remove <msgid>",		do_remove);
-	cmd_install("remove all",		do_remove);
-	cmd_install("resume envelope <evpid>",	do_resume_envelope);
-	cmd_install("resume envelope <msgid>",	do_resume_envelope);
-	cmd_install("resume envelope all",	do_resume_envelope);
-	cmd_install("resume mda",		do_resume_mda);
-	cmd_install("resume mta",		do_resume_mta);
-	cmd_install("resume route <routeid>",	do_resume_route);
-	cmd_install("resume smtp",		do_resume_smtp);
-	cmd_install("schedule <msgid>",		do_schedule);
-	cmd_install("schedule <evpid>",		do_schedule);
-	cmd_install("schedule all",		do_schedule);
-	cmd_install("show envelope <evpid>",	do_show_envelope);
-	cmd_install("show hoststats",		do_show_hoststats);
-	cmd_install("show message <msgid>",	do_show_message);
-	cmd_install("show message <evpid>",	do_show_message);
-	cmd_install("show queue",		do_show_queue);
-	cmd_install("show queue <msgid>",	do_show_queue);
-	cmd_install("show hosts",		do_show_hosts);
-	cmd_install("show relays",		do_show_relays);
-	cmd_install("show routes",		do_show_routes);
-	cmd_install("show stats",		do_show_stats);
-	cmd_install("show status",		do_show_status);
-	cmd_install("trace <str>",		do_trace);
-	cmd_install("uncorrupt <msgid>",	do_uncorrupt);
-	cmd_install("unprofile <str>",		do_unprofile);
-	cmd_install("untrace <str>",		do_untrace);
-	cmd_install("update table <str>",	do_update_table);
+	cmd_install("spf walk",			do_spf_walk);
 
 	if (strcmp(__progname, "mailq") == 0)
 		return cmd_run(2, argv_mailq);
@@ -1106,7 +1138,7 @@ sendmail_compat(int argc, char **argv)
 {
 	FILE	*offlinefp = NULL;
 	gid_t	 gid;
-	int	 i;
+	int	 i, r;
 
 	if (strcmp(__progname, "sendmail") == 0 ||
 	    strcmp(__progname, "send-mail") == 0) {
@@ -1125,24 +1157,36 @@ sendmail_compat(int argc, char **argv)
 		if (setresgid(gid, gid, gid) == -1)
 			err(1, "setresgid");
 
+#if HAVE_PLEDGE
 		/* we'll reduce further down the road */
 		if (pledge("stdio rpath wpath cpath tmppath flock "
 			"dns getpw recvfd", NULL) == -1)
 			err(1, "pledge");
+#endif
 
 		sendmail = 1;
 		exit(enqueue(argc, argv, offlinefp));
 	} else if (strcmp(__progname, "makemap") == 0)
 		exit(makemap(P_MAKEMAP, argc, argv));
-	else if (strcmp(__progname, "newaliases") == 0)
-		exit(makemap(P_NEWALIASES, argc, argv));
+	else if (strcmp(__progname, "newaliases") == 0) {
+		r = makemap(P_NEWALIASES, argc, argv);
+		/*
+		 * if server is available, notify of table update.
+		 * only makes sense for static tables AND if server is up.
+		 */
+		if (srv_connect()) {
+			srv_send(IMSG_CTL_UPDATE_TABLE, "aliases", strlen("aliases") + 1);
+			srv_check_result(0);
+		}
+		exit(r);
+	}
 }
 
 static void
 show_queue_envelope(struct envelope *e, int online)
 {
 	const char	*src = "?", *agent = "?";
-	char		 status[128], runstate[128];
+	char		 status[128], runstate[128], errline[LINE_MAX];
 
 	status[0] = '\0';
 
@@ -1187,6 +1231,8 @@ show_queue_envelope(struct envelope *e, int online)
 	else if (e->ss.ss_family == AF_INET6)
 		src = "inet6";
 
+	strnvis(errline, e->errorline, sizeof(errline), 0);
+
 	printf("%016"PRIx64
 	    "|%s|%s|%s|%s@%s|%s@%s|%s@%s"
 	    "|%zu|%zu|%zu|%zu|%s|%s\n",
@@ -1201,11 +1247,11 @@ show_queue_envelope(struct envelope *e, int online)
 	    e->dest.user, e->dest.domain,
 
 	    (size_t) e->creation,
-	    (size_t) (e->creation + e->expire),
+	    (size_t) (e->creation + e->ttl),
 	    (size_t) e->lasttry,
 	    (size_t) e->retry,
 	    runstate,
-	    e->errorline);
+	    errline);
 }
 
 static void
@@ -1268,9 +1314,7 @@ static void
 display(const char *s)
 {
 	FILE   *fp;
-#ifdef HAVE_GCM_CRYPTO
 	char   *key;
-#endif
 	int	gzipped;
 	char   *gzcat_argv0 = strrchr(PATH_GZCAT, '/') + 1;
 
@@ -1278,22 +1322,11 @@ display(const char *s)
 		err(1, "fopen");
 
 	if (is_encrypted_fp(fp)) {
-#ifdef HAVE_GCM_CRYPTO
 		int	i;
-		int	fd;
 		FILE   *ofp = NULL;
-		char	sfn[] = "/tmp/smtpd.XXXXXXXXXX";
 
-		if ((fd = mkstemp(sfn)) == -1 ||
-		    (ofp = fdopen(fd, "w+")) == NULL) {
-			int saved_errno = errno;
-			if (fd != -1) {
-				unlink(sfn);
-				close(fd);
-			}
-			errc(1, saved_errno, "mkstemp");
-		}
-		unlink(sfn);
+		if ((ofp = tmpfile()) == NULL)
+			err(1, "tmpfile");
 
 		for (i = 0; i < 3; i++) {
 			key = getpass("key> ");
@@ -1311,10 +1344,6 @@ display(const char *s)
 		fclose(fp);
 		fp = ofp;
 		fseek(fp, 0, SEEK_SET);
-#else
-	       	printf("GCM crypto not supported!\n");
-       		exit(1);
-#endif
 	}
 	gzipped = is_gzip_fp(fp);
 
