@@ -1,4 +1,4 @@
-/*	$OpenBSD: smtpd.h,v 1.516 2016/07/24 16:04:53 gilles Exp $	*/
+/*	$OpenBSD: smtpd.h,v 1.642 2019/11/03 23:58:51 gilles Exp $	*/
 
 /*
  * Copyright (c) 2008 Gilles Chehade <gilles@poolp.org>
@@ -28,12 +28,13 @@
 #define nitems(_a) (sizeof((_a)) / sizeof((_a)[0]))
 #endif
 
+#include <netinet/in.h>
+#include <netdb.h>
+#include <event.h>
+
 #include "smtpd-defines.h"
 #include "smtpd-api.h"
 #include "ioev.h"
-#include "iobuf.h"
-
-#include "rfc2822.h"
 
 #define CHECK_IMSG_DATA_SIZE(imsg, expected_sz) do {			\
 	if ((imsg)->hdr.len - IMSG_HEADER_SIZE != (expected_sz))	\
@@ -55,8 +56,6 @@
 
 #define MAX_HOPS_COUNT		 100
 #define	DEFAULT_MAX_BODY_SIZE	(35*1024*1024)
-#define	MAX_FILTER_NAME		 32
-#define	MAX_FILTER_ARGS		 255
 
 #define	EXPAND_BUFFER		 1024
 
@@ -74,7 +73,7 @@
 #ifndef SMTPD_NAME
 #define	SMTPD_NAME		 "OpenSMTPD"
 #endif
-#define	SMTPD_VERSION		 "6.0.2p1"
+#define	SMTPD_VERSION		 "6.6.2p1"
 #define SMTPD_SESSION_TIMEOUT	 300
 #define SMTPD_BACKLOG		 5
 
@@ -96,38 +95,44 @@
  * potentially dangerous and need to be escaped.
  */
 #define	MAILADDR_ALLOWED       	"!#$%&'*/?^`{|}~+-=_"
-#define	MAILADDR_ESCAPE		"!#$%&'*/?^`{|}~"
+#define	MAILADDR_ESCAPE		"!#$%&'*?`{|}~"
 
 
 #define F_STARTTLS		0x01
 #define F_SMTPS			0x02
-#define	F_TLS_OPTIONAL		0x04
 #define F_SSL		       (F_STARTTLS | F_SMTPS)
 #define F_AUTH			0x08
-#define	F_BACKUP		0x10	/* XXX - MUST BE SYNC-ED WITH RELAY_BACKUP */
 #define	F_STARTTLS_REQUIRE	0x20
 #define	F_AUTH_REQUIRE		0x40
-#define	F_LMTP			0x80
 #define	F_MASK_SOURCE		0x100
 #define	F_TLS_VERIFY		0x200
 #define	F_EXT_DSN		0x400
 #define	F_RECEIVEDAUTH		0x800
 #define	F_MASQUERADE		0x1000
+#define	F_FILTERED		0x2000
+#define	F_PROXY			0x4000
+
+#define RELAY_TLS_OPPORTUNISTIC	0
+#define RELAY_TLS_STARTTLS	1
+#define RELAY_TLS_SMTPS		2
+#define RELAY_TLS_NO		3
 
-/* must match F_* for mta */
-#define RELAY_STARTTLS		0x01
-#define RELAY_SMTPS		0x02
-#define	RELAY_TLS_OPTIONAL     	0x04
-#define RELAY_SSL		(RELAY_STARTTLS | RELAY_SMTPS)
 #define RELAY_AUTH		0x08
-#define RELAY_BACKUP		0x10	/* XXX - MUST BE SYNC-ED WITH F_BACKUP */
-#define RELAY_MX		0x20
 #define RELAY_LMTP		0x80
 #define	RELAY_TLS_VERIFY	0x200
 
 #define MTA_EXT_DSN		0x400
 
 
+#define P_NEWALIASES	0
+#define P_MAKEMAP	1
+
+#define	CERT_ERROR	-1
+#define	CERT_OK		 0
+#define	CERT_NOCA	 1
+#define	CERT_NOCERT	 2
+#define	CERT_INVALID	 3
+
 struct userinfo {
 	char username[SMTPD_VUSERNAME_SIZE];
 	char directory[PATH_MAX];
@@ -142,15 +147,10 @@ struct netaddr {
 
 struct relayhost {
 	uint16_t flags;
+	int tls;
 	char hostname[HOST_NAME_MAX+1];
 	uint16_t port;
-	char pki_name[HOST_NAME_MAX+1];
-	char ca_name[HOST_NAME_MAX+1];
-	char authtable[SMTPD_TABLENAME_SIZE];
 	char authlabel[PATH_MAX];
-	char sourcetable[PATH_MAX];
-	char heloname[HOST_NAME_MAX+1];
-	char helotable[PATH_MAX];
 };
 
 struct credentials {
@@ -181,13 +181,14 @@ union lookup {
 	struct mailaddr		 mailaddr;
 	struct addrname		 addrname;
 	struct maddrmap		*maddrmap;
+	char			 relayhost[LINE_MAX];
 };
 
 /*
  * Bump IMSG_VERSION whenever a change is made to enum imsg_type.
  * This will ensure that we can never use a wrong version of smtpctl with smtpd.
  */
-#define	IMSG_VERSION		14
+#define	IMSG_VERSION		16
 
 enum imsg_type {
 	IMSG_NONE,
@@ -221,17 +222,24 @@ enum imsg_type {
 	IMSG_CTL_REMOVE,
 	IMSG_CTL_SCHEDULE,
 	IMSG_CTL_SHOW_STATUS,
-	IMSG_CTL_SHUTDOWN,
 	IMSG_CTL_TRACE_DISABLE,
 	IMSG_CTL_TRACE_ENABLE,
 	IMSG_CTL_UPDATE_TABLE,
 	IMSG_CTL_VERBOSE,
 	IMSG_CTL_DISCOVER_EVPID,
 	IMSG_CTL_DISCOVER_MSGID,
-	IMSG_CTL_UNCORRUPT_MSGID,
 
 	IMSG_CTL_SMTP_SESSION,
 
+	IMSG_GETADDRINFO,
+	IMSG_GETADDRINFO_END,
+	IMSG_GETNAMEINFO,
+	IMSG_RES_QUERY,
+
+	IMSG_CERT_INIT,
+	IMSG_CERT_CERTIFICATE,
+	IMSG_CERT_VERIFY,
+
 	IMSG_SETUP_KEY,
 	IMSG_SETUP_PEER,
 	IMSG_SETUP_DONE,
@@ -279,8 +287,6 @@ enum imsg_type {
 	IMSG_MDA_KILL,
 	IMSG_MDA_OPEN_MESSAGE,
 
-	IMSG_MFA_SMTP_RESPONSE,
-
 	IMSG_MTA_DELIVERY_OK,
 	IMSG_MTA_DELIVERY_TEMPFAIL,
 	IMSG_MTA_DELIVERY_PERMFAIL,
@@ -288,19 +294,15 @@ enum imsg_type {
 	IMSG_MTA_DELIVERY_HOLD,
 	IMSG_MTA_DNS_HOST,
 	IMSG_MTA_DNS_HOST_END,
-	IMSG_MTA_DNS_PTR,
 	IMSG_MTA_DNS_MX,
 	IMSG_MTA_DNS_MX_PREFERENCE,
 	IMSG_MTA_HOLDQ_RELEASE,
 	IMSG_MTA_LOOKUP_CREDENTIALS,
 	IMSG_MTA_LOOKUP_SOURCE,
 	IMSG_MTA_LOOKUP_HELO,
+	IMSG_MTA_LOOKUP_SMARTHOST,
 	IMSG_MTA_OPEN_MESSAGE,
 	IMSG_MTA_SCHEDULE,
-	IMSG_MTA_TLS_INIT,
-	IMSG_MTA_TLS_VERIFY_CERT,
-	IMSG_MTA_TLS_VERIFY_CHAIN,
-	IMSG_MTA_TLS_VERIFY,
 
 	IMSG_SCHED_ENVELOPE_BOUNCE,
 	IMSG_SCHED_ENVELOPE_DELIVER,
@@ -310,7 +312,6 @@ enum imsg_type {
 	IMSG_SCHED_ENVELOPE_TRANSFER,
 
 	IMSG_SMTP_AUTHENTICATE,
-	IMSG_SMTP_DNS_PTR,
 	IMSG_SMTP_MESSAGE_COMMIT,
 	IMSG_SMTP_MESSAGE_CREATE,
 	IMSG_SMTP_MESSAGE_ROLLBACK,
@@ -318,10 +319,6 @@ enum imsg_type {
 	IMSG_SMTP_CHECK_SENDER,
 	IMSG_SMTP_EXPAND_RCPT,
 	IMSG_SMTP_LOOKUP_HELO,
-	IMSG_SMTP_TLS_INIT,
-	IMSG_SMTP_TLS_VERIFY_CERT,
-	IMSG_SMTP_TLS_VERIFY_CHAIN,
-	IMSG_SMTP_TLS_VERIFY,
 
 	IMSG_SMTP_REQ_CONNECT,
 	IMSG_SMTP_REQ_HELO,
@@ -334,8 +331,37 @@ enum imsg_type {
 	IMSG_SMTP_EVENT_ROLLBACK,
 	IMSG_SMTP_EVENT_DISCONNECT,
 
-	IMSG_CA_PRIVENC,
-	IMSG_CA_PRIVDEC
+	IMSG_LKA_PROCESSOR_FORK,
+	IMSG_LKA_PROCESSOR_ERRFD,
+
+	IMSG_REPORT_SMTP_LINK_CONNECT,
+	IMSG_REPORT_SMTP_LINK_DISCONNECT,
+	IMSG_REPORT_SMTP_LINK_GREETING,
+	IMSG_REPORT_SMTP_LINK_IDENTIFY,
+	IMSG_REPORT_SMTP_LINK_TLS,
+	IMSG_REPORT_SMTP_LINK_AUTH,
+	IMSG_REPORT_SMTP_TX_RESET,
+	IMSG_REPORT_SMTP_TX_BEGIN,
+	IMSG_REPORT_SMTP_TX_MAIL,
+	IMSG_REPORT_SMTP_TX_RCPT,
+	IMSG_REPORT_SMTP_TX_ENVELOPE,
+	IMSG_REPORT_SMTP_TX_DATA,
+	IMSG_REPORT_SMTP_TX_COMMIT,
+	IMSG_REPORT_SMTP_TX_ROLLBACK,
+	IMSG_REPORT_SMTP_PROTOCOL_CLIENT,
+	IMSG_REPORT_SMTP_PROTOCOL_SERVER,
+	IMSG_REPORT_SMTP_FILTER_RESPONSE,
+	IMSG_REPORT_SMTP_TIMEOUT,
+
+	IMSG_FILTER_SMTP_BEGIN,
+	IMSG_FILTER_SMTP_END,
+	IMSG_FILTER_SMTP_PROTOCOL,
+	IMSG_FILTER_SMTP_DATA_BEGIN,
+	IMSG_FILTER_SMTP_DATA_END,
+
+	IMSG_CA_RSA_PRIVENC,
+	IMSG_CA_RSA_PRIVDEC,
+	IMSG_CA_ECDSA_SIGN,
 };
 
 enum smtp_proc_type {
@@ -346,8 +372,7 @@ enum smtp_proc_type {
 	PROC_SCHEDULER,
 	PROC_PONY,
 	PROC_CA,
-
-	PROC_FILTER,
+	PROC_PROCESSOR,
 	PROC_CLIENT,
 };
 
@@ -363,94 +388,28 @@ struct table {
 	enum table_type			 t_type;
 	char				 t_config[PATH_MAX];
 
-	struct dict			 t_dict;
-
 	void				*t_handle;
 	struct table_backend		*t_backend;
-	void				*t_iter;
 };
 
 struct table_backend {
+	const char *name;
 	const unsigned int	services;
 	int	(*config)(struct table *);
-	void   *(*open)(struct table *);
+	int	(*add)(struct table *, const char *, const char *);
+	void	(*dump)(struct table *);
+	int	(*open)(struct table *);
 	int	(*update)(struct table *);
-	void	(*close)(void *);
-	int	(*lookup)(void *, struct dict *, const char *, enum table_service, union lookup *);
-	int	(*fetch)(void *, struct dict *, enum table_service, union lookup *);
+	void	(*close)(struct table *);
+	int	(*lookup)(struct table *, enum table_service, const char *, char **);
+	int	(*fetch)(struct table *, enum table_service, char **);
 };
 
 
-enum dest_type {
-	DEST_DOM,
-	DEST_VDOM
-};
-
-enum action_type {
-	A_NONE,
-	A_RELAY,
-	A_RELAYVIA,
-	A_MAILDIR,
-	A_MBOX,
-	A_FILENAME,
-	A_MDA,
-	A_LMTP
-};
-
-enum decision {
-	R_REJECT,
-	R_ACCEPT
-};
-
-struct rule {
-	TAILQ_ENTRY(rule)		r_entry;
-	enum decision			r_decision;
-	uint8_t				r_nottag;
-	char				r_tag[SMTPD_TAG_SIZE];
-
-	uint8_t				r_notsources;
-	struct table		       *r_sources;
-
-	uint8_t				r_notsenders;
-	struct table		       *r_senders;
-
-	uint8_t				r_notrecipients;
-	struct table		       *r_recipients;
-
-	uint8_t				r_notdestination;
-	enum dest_type			r_desttype;
-	struct table		       *r_destination;
-
-	enum action_type		r_action;
-	union rule_dest {
-		char			buffer[EXPAND_BUFFER];
-		struct relayhost	relayhost;
-	}				r_value;
-
-	struct mailaddr		       *r_as;
-	struct table		       *r_mapping;
-	struct table		       *r_userbase;
-	time_t				r_qexpire;
-	uint8_t				r_forwardonly;
-	char				r_delivery_user[LINE_MAX];
-};
-
-struct delivery_mda {
-	enum action_type	method;
-	char			usertable[SMTPD_TABLENAME_SIZE];
-	char			username[SMTPD_VUSERNAME_SIZE];
-	char			buffer[EXPAND_BUFFER];
-	char			delivery_user[SMTPD_VUSERNAME_SIZE];
-};
-
-struct delivery_mta {
-	struct relayhost	relay;
-};
-
 enum bounce_type {
-	B_ERROR,
-	B_WARNING,
-	B_DSN
+	B_FAILED,
+	B_DELAYED,
+	B_DELIVERED
 };
 
 enum dsn_ret {
@@ -461,7 +420,7 @@ enum dsn_ret {
 struct delivery_bounce {
 	enum bounce_type	type;
 	time_t			delay;
-	time_t			expire;
+	time_t			ttl;
 	enum dsn_ret		dsn_ret;
         int			mta_without_dsn;
 };
@@ -474,7 +433,25 @@ enum expand_type {
 	EXPAND_INCLUDE,
 	EXPAND_ADDRESS,
 	EXPAND_ERROR,
-	EXPAND_MAILDIR
+};
+
+enum filter_phase {
+	FILTER_CONNECT,
+	FILTER_HELO,
+	FILTER_EHLO,
+	FILTER_STARTTLS,
+	FILTER_AUTH,
+	FILTER_MAIL_FROM,
+	FILTER_RCPT_TO,
+	FILTER_DATA,
+	FILTER_DATA_LINE,
+	FILTER_RSET,
+	FILTER_QUIT,
+	FILTER_NOOP,
+	FILTER_HELP,
+	FILTER_WIZ,
+	FILTER_COMMIT,
+	FILTER_PHASES_COUNT     /* must be last */
 };
 
 struct expandnode {
@@ -482,12 +459,11 @@ struct expandnode {
 	TAILQ_ENTRY(expandnode)	tq_entry;
 	enum expand_type	type;
 	int			sameuser;
-	int			alias;
+	int			realuser;
+	int			forwarded;
 	struct rule	       *rule;
 	struct expandnode      *parent;
 	unsigned int		depth;
-	struct table   	       *mapping;
-	struct table   	       *userbase;
 	union {
 		/*
 		 * user field handles both expansion user and system user
@@ -497,12 +473,12 @@ struct expandnode {
 		char		buffer[EXPAND_BUFFER];
 		struct mailaddr	mailaddr;
 	}			u;
+	char		subaddress[SMTPD_SUBADDRESS_SIZE];
 };
 
 struct expand {
 	RB_HEAD(expandtree, expandnode)	 tree;
 	TAILQ_HEAD(xnodes, expandnode)	*queue;
-	int				 alias;
 	size_t				 nb_nodes;
 	struct rule			*rule;
 	struct expandnode		*parent;
@@ -524,10 +500,12 @@ struct maddrmap {
 
 #define	DSN_ENVID_LEN	100
 
-#define	SMTPD_ENVELOPE_VERSION		2
+#define	SMTPD_ENVELOPE_VERSION		3
 struct envelope {
 	TAILQ_ENTRY(envelope)		entry;
 
+	char				dispatcher[HOST_NAME_MAX+1];
+
 	char				tag[SMTPD_TAG_SIZE];
 
 	uint32_t			version;
@@ -544,16 +522,18 @@ struct envelope {
 	struct mailaddr			rcpt;
 	struct mailaddr			dest;
 
+	char				mda_user[SMTPD_VUSERNAME_SIZE];
+	char				mda_subaddress[SMTPD_SUBADDRESS_SIZE];
+	char				mda_exec[LINE_MAX];
+
 	enum delivery_type		type;
 	union {
-		struct delivery_mda	mda;
-		struct delivery_mta	mta;
 		struct delivery_bounce	bounce;
 	}				agent;
 
 	uint16_t			retry;
 	time_t				creation;
-	time_t				expire;
+	time_t				ttl;
 	time_t				lasttry;
 	time_t				nexttry;
 	time_t				lastbounce;
@@ -574,10 +554,10 @@ struct listener {
 	in_port_t		 port;
 	struct timeval		 timeout;
 	struct event		 ev;
+	char			 filter_name[PATH_MAX];
 	char			 pki_name[PATH_MAX];
 	char			 ca_name[PATH_MAX];
 	char			 tag[SMTPD_TAG_SIZE];
-	char			 filter[PATH_MAX];
 	char			 authtable[LINE_MAX];
 	char			 hostname[HOST_NAME_MAX+1];
 	char			 hostnametable[PATH_MAX];
@@ -596,7 +576,7 @@ struct smtpd {
 #define SMTPD_OPT_NOACTION		0x00000002
 	uint32_t			sc_opts;
 
-#define SMTPD_EXITING			0x00000001
+#define SMTPD_EXITING			0x00000001 /* unused */
 #define SMTPD_MDA_PAUSED		0x00000002
 #define SMTPD_MTA_PAUSED		0x00000004
 #define SMTPD_SMTP_PAUSED		0x00000008
@@ -616,6 +596,7 @@ struct smtpd {
 	size_t				sc_session_max_rcpt;
 	size_t				sc_session_max_mails;
 
+	struct dict		       *sc_mda_wrappers;
 	size_t				sc_mda_max_session;
 	size_t				sc_mda_max_user_session;
 	size_t				sc_mda_task_hiwat;
@@ -629,7 +610,9 @@ struct smtpd {
 	size_t				sc_scheduler_max_msg_batch_size;
 	size_t				sc_scheduler_max_schedule;
 
-	int				sc_qexpire;
+	struct dict		       *sc_processors_dict;
+
+	int				sc_ttl;
 #define MAX_BOUNCE_WARN			4
 	time_t				sc_bounce_warn[MAX_BOUNCE_WARN];
 	char				sc_hostname[HOST_NAME_MAX+1];
@@ -645,6 +628,11 @@ struct smtpd {
 
 	TAILQ_HEAD(rulelist, rule)		*sc_rules;
 
+
+	struct dict				*sc_filters_dict;
+	struct dict				*sc_dispatchers;
+	struct dispatcher			*sc_dispatcher_bounce;
+
 	struct dict			       *sc_ca_dict;
 	struct dict			       *sc_pki_dict;
 	struct dict			       *sc_ssl_dict;
@@ -653,12 +641,13 @@ struct smtpd {
 
 	struct dict			       *sc_limits_dict;
 
-	struct dict				sc_filters;
-	uint32_t				filtermask;
-
-	char					sc_enqueue_filter[PATH_MAX];
-
 	char				       *sc_tls_ciphers;
+
+	char				       *sc_subaddressing_delim;
+
+	char				       *sc_srs_key;
+	char				       *sc_srs_key_backup;
+	int				        sc_srs_ttl;
 };
 
 #define	TRACE_DEBUG	0x0001
@@ -680,7 +669,6 @@ struct smtpd {
 #define PROFILE_TOSTAT	0x0001
 #define PROFILE_IMSG	0x0002
 #define PROFILE_QUEUE	0x0004
-#define PROFILE_BUFFERS	0x0008
 
 struct forward_req {
 	uint64_t			id;
@@ -693,25 +681,18 @@ struct forward_req {
 };
 
 struct deliver {
-	char			to[EXPAND_BUFFER];
-	char			from[SMTPD_MAXMAILADDRSIZE];
-	char			dest[SMTPD_MAXMAILADDRSIZE];
-	char			user[SMTPD_VUSERNAME_SIZE];
-	short			mode;
+	char			dispatcher[EXPAND_BUFFER];
+
+	struct mailaddr		sender;
+	struct mailaddr		rcpt;
+	struct mailaddr		dest;
+
+	char			mda_subaddress[SMTPD_SUBADDRESS_SIZE];
+	char			mda_exec[LINE_MAX];
 
 	struct userinfo		userinfo;
 };
 
-#define MAX_FILTER_PER_CHAIN	16
-struct filter_conf {
-	int		 chain;
-	int		 done;
-	int		 argc;
-	char		*name;
-	char		*argv[MAX_FILTER_ARGS + 1];
-	char		*path;
-};
-
 struct mta_host {
 	SPLAY_ENTRY(mta_host)	 entry;
 	struct sockaddr		*sa;
@@ -728,13 +709,14 @@ struct mta_host {
 struct mta_mx {
 	TAILQ_ENTRY(mta_mx)	 entry;
 	struct mta_host		*host;
+	char			*mxname;
 	int			 preference;
 };
 
 struct mta_domain {
 	SPLAY_ENTRY(mta_domain)	 entry;
 	char			*name;
-	int			 flags;
+	int			 as_host;
 	TAILQ_HEAD(, mta_mx)	 mxs;
 	int			 mxstatus;
 	int			 refcount;
@@ -836,8 +818,10 @@ struct mta_relay {
 	SPLAY_ENTRY(mta_relay)	 entry;
 	uint64_t		 id;
 
+	struct dispatcher	*dispatcher;
 	struct mta_domain	*domain;
 	struct mta_limits	*limits;
+	int			 tls;
 	int			 flags;
 	char			*backupname;
 	int			 backuppref;
@@ -850,6 +834,7 @@ struct mta_relay {
 	char			*helotable;
 	char			*heloname;
 	char			*secret;
+	int			 srs;
 
 	int			 state;
 	size_t			 ntask;
@@ -869,7 +854,8 @@ struct mta_relay {
 #define RELAY_WAIT_LIMITS	0x08
 #define RELAY_WAIT_SOURCE	0x10
 #define RELAY_WAIT_CONNECTOR	0x20
-#define RELAY_WAITMASK		0x3f
+#define RELAY_WAIT_SMARTHOST	0x40
+#define RELAY_WAITMASK		0x7f
 	int			 status;
 
 	int			 refcount;
@@ -883,6 +869,7 @@ struct mta_envelope {
 	uint64_t			 id;
 	uint64_t			 session;
 	time_t				 creation;
+	char				*smtpname;
 	char				*dest;
 	char				*rcpt;
 	struct mta_task			*task;
@@ -928,13 +915,6 @@ struct auth_backend {
 	int	(*authenticate)(char *, char *);
 };
 
-
-/* delivery_backend */
-struct delivery_backend {
-	int	allow_root;
-	void	(*open)(struct deliver *);
-};
-
 struct scheduler_backend {
 	int	(*init)(const char *);
 
@@ -1031,13 +1011,6 @@ struct mproc {
 	short		 events;
 	struct event	 ev;
 	void		*data;
-
-	off_t		 msg_in;
-	off_t		 msg_out;
-	off_t		 bytes_in;
-	off_t		 bytes_out;
-	size_t		 bytes_queued;
-	size_t		 bytes_queued_max;
 };
 
 struct msg {
@@ -1047,7 +1020,7 @@ struct msg {
 
 extern enum smtp_proc_type	smtpd_process;
 
-extern int verbose;
+extern int tracing;
 extern int foreground_log;
 extern int profiling;
 
@@ -1062,16 +1035,6 @@ extern struct mproc *p_ca;
 extern struct smtpd	*env;
 extern void (*imsg_callback)(struct mproc *, struct imsg *);
 
-struct imsgproc {
-	pid_t			pid;
-	struct event		ev;
-	struct imsgbuf	       *ibuf;
-	char		       *path;
-	char		       *name;
-	void		      (*cb)(struct imsg *, void *);
-	void		       *cb_arg;
-};
-
 /* inter-process structures */
 
 struct bounce_req_msg {
@@ -1094,40 +1057,90 @@ enum lka_resp_status {
 	LKA_PERMFAIL
 };
 
+struct processor {
+	const char		       *command;
+	const char		       *user;
+	const char		       *group;
+	const char		       *chroot;
+	int				errfd;
+};
+
+enum filter_type {
+	FILTER_TYPE_BUILTIN,
+	FILTER_TYPE_PROC,
+	FILTER_TYPE_CHAIN,
+};
+
+struct filter_config {
+	char			       *name;
+	enum filter_type		filter_type;
+	enum filter_phase               phase;
+	char                           *reject;
+	char                           *disconnect;
+	char                           *rewrite;
+	char                           *report;
+	uint8_t				junk;
+	char                           *proc;
+
+	const char		      **chain;
+	size_t				chain_size;
+	struct dict			chain_procs;
+
+	int8_t				not_fcrdns;
+	int8_t				fcrdns;
+
+	int8_t				not_rdns;
+	int8_t				rdns;
+
+	int8_t                          not_rdns_table;
+	struct table                   *rdns_table;
+
+	int8_t                          not_rdns_regex;
+	struct table                   *rdns_regex;
+
+	int8_t                          not_src_table;
+	struct table                   *src_table;
+
+	int8_t                          not_src_regex;
+	struct table                   *src_regex;
+
+	int8_t                          not_helo_table;
+	struct table                   *helo_table;
+
+	int8_t                          not_helo_regex;
+	struct table                   *helo_regex;
+
+	int8_t                          not_mail_from_table;
+	struct table                   *mail_from_table;
+
+	int8_t                          not_mail_from_regex;
+	struct table                   *mail_from_regex;
+
+	int8_t                          not_rcpt_to_table;
+	struct table                   *rcpt_to_table;
+
+	int8_t                          not_rcpt_to_regex;
+	struct table                   *rcpt_to_regex;
+
+};
+
+enum filter_status {
+	FILTER_PROCEED,
+	FILTER_REWRITE,
+	FILTER_REJECT,
+	FILTER_DISCONNECT,
+	FILTER_JUNK,
+};
+
 enum ca_resp_status {
 	CA_OK,
 	CA_FAIL
 };
 
-struct ca_cert_req_msg {
-	uint64_t		reqid;
-	char			name[HOST_NAME_MAX+1];
-	int			fallback;
-};
-
-struct ca_cert_resp_msg {
-	uint64_t		reqid;
-	enum ca_resp_status	status;
-	char			name[HOST_NAME_MAX+1];
-	char		       *cert;
-	off_t			cert_len;
-};
-
-struct ca_vrfy_req_msg {
-	uint64_t		reqid;
-	char			name[HOST_NAME_MAX+1];
-	int			fallback;
-	unsigned char  	       *cert;
-	off_t			cert_len;
-	size_t			n_chain;
-	size_t			chain_offset;
-	unsigned char	      **chain_cert;
-	off_t		       *chain_cert_len;
-};
-
-struct ca_vrfy_resp_msg {
-	uint64_t		reqid;
-	enum ca_resp_status	status;
+enum mda_resp_status {
+	MDA_OK,
+	MDA_TEMPFAIL,
+	MDA_PERMFAIL
 };
 
 struct msg_walkinfo {
@@ -1139,6 +1152,106 @@ struct msg_walkinfo {
 	int		 done;
 };
 
+
+enum dispatcher_type {
+	DISPATCHER_LOCAL,
+	DISPATCHER_REMOTE,
+	DISPATCHER_BOUNCE,
+};
+
+struct dispatcher_local {
+	uint8_t requires_root;	/* only for MBOX */
+
+	uint8_t	expand_only;
+	uint8_t	forward_only;
+
+	char	*mda_wrapper;
+	char	*command;
+
+	char	*table_alias;
+	char	*table_virtual;
+	char	*table_userbase;
+
+	char	*user;
+};
+
+struct dispatcher_remote {
+	char	*helo;
+	char	*helo_source;
+
+	char	*source;
+
+	char	*ca;
+	char	*pki;
+
+	char	*mail_from;
+
+	char	*smarthost;
+	char	*auth;
+	int	 tls_required;
+	int	 tls_noverify;
+
+	int	 backup;
+	char	*backupmx;
+
+	int	 srs;
+};
+
+struct dispatcher_bounce {
+};
+
+struct dispatcher {
+	enum dispatcher_type			type;
+	union dispatcher_agent {
+		struct dispatcher_local		local;
+		struct dispatcher_remote  	remote;
+		struct dispatcher_bounce  	bounce;
+	} u;
+
+	time_t	ttl;
+};
+
+struct rule {
+	TAILQ_ENTRY(rule)	r_entry;
+
+	uint8_t	reject;
+
+	int8_t	flag_tag;
+	int8_t	flag_from;
+	int8_t	flag_for;
+	int8_t	flag_from_rdns;
+	int8_t	flag_from_socket;
+
+	int8_t	flag_tag_regex;
+	int8_t	flag_from_regex;
+	int8_t	flag_for_regex;
+
+	int8_t	flag_smtp_helo;
+	int8_t	flag_smtp_starttls;
+	int8_t	flag_smtp_auth;
+	int8_t	flag_smtp_mail_from;
+	int8_t	flag_smtp_rcpt_to;
+
+	int8_t	flag_smtp_helo_regex;
+	int8_t	flag_smtp_starttls_regex;
+	int8_t	flag_smtp_auth_regex;
+	int8_t	flag_smtp_mail_from_regex;
+	int8_t	flag_smtp_rcpt_to_regex;
+
+
+	char	*table_tag;
+	char	*table_from;
+	char	*table_for;
+
+	char	*table_smtp_helo;
+	char	*table_smtp_auth;
+	char	*table_smtp_mail_from;
+	char	*table_smtp_rcpt_to;
+
+	char	*dispatcher;
+};
+
+
 /* aliases.c */
 int aliases_get(struct expand *, const char *);
 int aliases_virtual_get(struct expand *, const struct mailaddr *);
@@ -1161,6 +1274,15 @@ void	 ca_imsg(struct mproc *, struct imsg *);
 void	 ca_init(void);
 void	 ca_engine_init(void);
 
+
+/* cert.c */
+int cert_init(const char *, int,
+    void (*)(void *, int, const char *, const void *, size_t), void *);
+int cert_verify(const void *, const char *, int, void (*)(void *, int), void *);
+void cert_dispatch_request(struct mproc *, struct imsg *);
+void cert_dispatch_result(struct mproc *, struct imsg *);
+
+
 /* compress_backend.c */
 struct compress_backend *compress_backend_lookup(const char *);
 size_t	compress_chunk(void *, size_t, void *, size_t);
@@ -1174,11 +1296,12 @@ int	uncompress_file(FILE *, FILE *);
 #define PURGE_RULES		0x04
 #define PURGE_PKI		0x08
 #define PURGE_PKI_KEYS		0x10
-#define PURGE_EVERYTHING	0x0f
+#define PURGE_DISPATCHERS	0x20
+#define PURGE_EVERYTHING	0xff
+struct smtpd *config_default(void);
 void purge_config(uint8_t);
 void config_process(enum smtp_proc_type);
 void config_peer(enum smtp_proc_type);
-void config_done(void);
 
 
 /* control.c */
@@ -1194,10 +1317,6 @@ size_t	crypto_encrypt_buffer(const char *, size_t, char *, size_t);
 size_t	crypto_decrypt_buffer(const char *, size_t, char *, size_t);
 
 
-/* delivery.c */
-struct delivery_backend *delivery_backend_lookup(enum action_type);
-
-
 /* dns.c */
 void dns_imsg(struct mproc *, struct imsg *);
 
@@ -1225,32 +1344,10 @@ int expand_to_text(struct expand *, char *, size_t);
 RB_PROTOTYPE(expandtree, expandnode, nodes, expand_cmp);
 
 
-/* filter.c */
-void filter_postfork(void);
-void filter_configure(void);
-void filter_connect(uint64_t, const struct sockaddr *,
-    const struct sockaddr *, const char *, const char *);
-void filter_mailaddr(uint64_t, int, const struct mailaddr *);
-void filter_line(uint64_t, int, const char *);
-void filter_eom(uint64_t, int, size_t);
-void filter_event(uint64_t, int);
-void filter_build_fd_chain(uint64_t, int);
-
-
 /* forward.c */
 int forwards_get(int, struct expand *);
 
 
-/* imsgproc.c */
-void imsgproc_init(void);
-struct imsgproc *imsgproc_fork(const char *, const char *,
-    void (*)(struct imsg *, void *), void *);
-void imsgproc_set_read(struct imsgproc *);
-void imsgproc_set_write(struct imsgproc *);
-void imsgproc_set_read_write(struct imsgproc *);
-void imsgproc_reset_callback(struct imsgproc *, void (*)(struct imsg *, void *), void *);
-
-
 /* limit.c */
 void limit_mta_set_defaults(struct mta_limits *);
 int limit_mta_set(struct mta_limits *, const char*, int64_t);
@@ -1260,6 +1357,55 @@ int limit_mta_set(struct mta_limits *, const char*, int64_t);
 int lka(void);
 
 
+/* lka_proc.c */
+int lka_proc_ready(void);
+void lka_proc_forked(const char *, int);
+void lka_proc_errfd(const char *, int);
+struct io *lka_proc_get_io(const char *);
+
+
+/* lka_report.c */
+void lka_report_init(void);
+void lka_report_register_hook(const char *, const char *);
+void lka_report_smtp_link_connect(const char *, struct timeval *, uint64_t, const char *, int,
+    const struct sockaddr_storage *, const struct sockaddr_storage *);
+void lka_report_smtp_link_disconnect(const char *, struct timeval *, uint64_t);
+void lka_report_smtp_link_greeting(const char *, uint64_t, struct timeval *,
+    const char *);
+void lka_report_smtp_link_identify(const char *, struct timeval *, uint64_t, const char *, const char *);
+void lka_report_smtp_link_tls(const char *, struct timeval *, uint64_t, const char *);
+void lka_report_smtp_link_auth(const char *, struct timeval *, uint64_t, const char *, const char *);
+void lka_report_smtp_tx_reset(const char *, struct timeval *, uint64_t, uint32_t);
+void lka_report_smtp_tx_begin(const char *, struct timeval *, uint64_t, uint32_t);
+void lka_report_smtp_tx_mail(const char *, struct timeval *, uint64_t, uint32_t, const char *, int);
+void lka_report_smtp_tx_rcpt(const char *, struct timeval *, uint64_t, uint32_t, const char *, int);
+void lka_report_smtp_tx_envelope(const char *, struct timeval *, uint64_t, uint32_t, uint64_t);
+void lka_report_smtp_tx_commit(const char *, struct timeval *, uint64_t, uint32_t, size_t);
+void lka_report_smtp_tx_data(const char *, struct timeval *, uint64_t, uint32_t, int);
+void lka_report_smtp_tx_rollback(const char *, struct timeval *, uint64_t, uint32_t);
+void lka_report_smtp_protocol_client(const char *, struct timeval *, uint64_t, const char *);
+void lka_report_smtp_protocol_server(const char *, struct timeval *, uint64_t, const char *);
+void lka_report_smtp_filter_response(const char *, struct timeval *, uint64_t,
+    int, int, const char *);
+void lka_report_smtp_timeout(const char *, struct timeval *, uint64_t);
+void lka_report_filter_report(uint64_t, const char *, int, const char *,
+    struct timeval *, const char *);
+void lka_report_proc(const char *, const char *);
+
+
+/* lka_filter.c */
+void lka_filter_init(void);
+void lka_filter_register_hook(const char *, const char *);
+void lka_filter_ready(void);
+int lka_filter_proc_in_session(uint64_t, const char *);
+void lka_filter_begin(uint64_t, const char *, const struct sockaddr_storage *, const struct sockaddr_storage *, const char *, int);
+void lka_filter_end(uint64_t);
+void lka_filter_protocol(uint64_t, enum filter_phase, const char *);
+void lka_filter_data_begin(uint64_t);
+void lka_filter_data_end(uint64_t);
+int lka_filter_response(uint64_t, const char *, const char *);
+
+
 /* lka_session.c */
 void lka_session(uint64_t, struct envelope *);
 void lka_session_forward_reply(struct forward_req *, int);
@@ -1276,8 +1422,17 @@ void mda_postprivdrop(void);
 void mda_imsg(struct mproc *, struct imsg *);
 
 
+/* mda_unpriv.c */
+void mda_unpriv(struct dispatcher *, struct deliver *, const char *, const char *);
+
+
+/* mda_variables.c */
+ssize_t mda_expand_format(char *, size_t, const struct deliver *,
+    const struct userinfo *, const char *);
+
+
 /* makemap.c */
-int makemap(int, char **);
+int makemap(int, int, char **);
 
 
 /* mailaddr.c */
@@ -1304,6 +1459,7 @@ void m_add_int(struct mproc *, int);
 void m_add_u32(struct mproc *, uint32_t);
 void m_add_size(struct mproc *, size_t);
 void m_add_time(struct mproc *, time_t);
+void m_add_timeval(struct mproc *, struct timeval *tv);
 void m_add_string(struct mproc *, const char *);
 void m_add_data(struct mproc *, const void *, size_t);
 void m_add_evpid(struct mproc *, uint64_t);
@@ -1323,6 +1479,7 @@ void m_get_int(struct msg *, int *);
 void m_get_size(struct msg *, size_t *);
 void m_get_u32(struct msg *, uint32_t *);
 void m_get_time(struct msg *, time_t *);
+void m_get_timeval(struct msg *, struct timeval *);
 void m_get_string(struct msg *, const char **);
 void m_get_data(struct msg *, const void **, size_t *);
 void m_get_evpid(struct msg *, uint64_t *);
@@ -1352,7 +1509,7 @@ const char *mta_relay_to_text(struct mta_relay *);
 
 
 /* mta_session.c */
-void mta_session(struct mta_relay *, struct mta_route *);
+void mta_session(struct mta_relay *, struct mta_route *, const char *);
 void mta_session_imsg(struct mproc *, struct imsg *);
 
 
@@ -1363,7 +1520,6 @@ int cmdline_symset(char *);
 
 /* queue.c */
 int queue(void);
-void queue_flow_control(void);
 
 
 /* queue_backend.c */
@@ -1376,8 +1532,6 @@ int queue_message_delete(uint32_t);
 int queue_message_commit(uint32_t);
 int queue_message_fd_r(uint32_t);
 int queue_message_fd_rw(uint32_t);
-int queue_message_corrupt(uint32_t);
-int queue_message_uncorrupt(uint32_t);
 int queue_envelope_create(struct envelope *);
 int queue_envelope_delete(uint64_t);
 int queue_envelope_load(uint64_t, struct envelope *);
@@ -1386,6 +1540,28 @@ int queue_envelope_walk(struct envelope *);
 int queue_message_walk(struct envelope *, uint32_t, int *, void **);
 
 
+/* report_smtp.c */
+void report_smtp_link_connect(const char *, uint64_t, const char *, int,
+    const struct sockaddr_storage *, const struct sockaddr_storage *);
+void report_smtp_link_disconnect(const char *, uint64_t);
+void report_smtp_link_greeting(const char *, uint64_t, const char *);
+void report_smtp_link_identify(const char *, uint64_t, const char *, const char *);
+void report_smtp_link_tls(const char *, uint64_t, const char *);
+void report_smtp_link_auth(const char *, uint64_t, const char *, const char *);
+void report_smtp_tx_reset(const char *, uint64_t, uint32_t);
+void report_smtp_tx_begin(const char *, uint64_t, uint32_t);
+void report_smtp_tx_mail(const char *, uint64_t, uint32_t, const char *, int);
+void report_smtp_tx_rcpt(const char *, uint64_t, uint32_t, const char *, int);
+void report_smtp_tx_envelope(const char *, uint64_t, uint32_t, uint64_t);
+void report_smtp_tx_data(const char *, uint64_t, uint32_t, int);
+void report_smtp_tx_commit(const char *, uint64_t, uint32_t, size_t);
+void report_smtp_tx_rollback(const char *, uint64_t, uint32_t);
+void report_smtp_protocol_client(const char *, uint64_t, const char *);
+void report_smtp_protocol_server(const char *, uint64_t, const char *);
+void report_smtp_filter_response(const char *, uint64_t, int, int, const char *);
+void report_smtp_timeout(const char *, uint64_t);
+
+
 /* ruleset.c */
 struct rule *ruleset_match(const struct envelope *);
 
@@ -1404,6 +1580,17 @@ int pony(void);
 void pony_imsg(struct mproc *, struct imsg *);
 
 
+/* resolver.c */
+void resolver_getaddrinfo(const char *, const char *, const struct addrinfo *,
+    void(*)(void *, int, struct addrinfo*), void *);
+void resolver_getnameinfo(const struct sockaddr *, int,
+    void(*)(void *, int, const char *, const char *), void *);
+void resolver_res_query(const char *, int, int,
+    void (*cb)(void *, int, int, int, const void *, int), void *);
+void resolver_dispatch_request(struct mproc *, struct imsg *);
+void resolver_dispatch_result(struct mproc *, struct imsg *);
+
+
 /* smtp.c */
 void smtp_postfork(void);
 void smtp_postprivdrop(void);
@@ -1414,10 +1601,14 @@ void smtp_collect(void);
 
 /* smtp_session.c */
 int smtp_session(struct listener *, int, const struct sockaddr_storage *,
-    const char *);
+    const char *, struct io *);
 void smtp_session_imsg(struct mproc *, struct imsg *);
-void smtp_filter_response(uint64_t, int, int, uint32_t, const char *);
-void smtp_filter_fd(uint64_t, int);
+
+
+/* smtpf_session.c */
+int smtpf_session(struct listener *, int, const struct sockaddr_storage *,
+    const char *);
+void smtpf_session_imsg(struct mproc *, struct imsg *);
 
 
 /* smtpd.c */
@@ -1429,6 +1620,11 @@ void log_imsg(int, int, struct imsg *);
 int fork_proc_backend(const char *, const char *, const char *);
 
 
+/* srs.c */
+const char *srs_encode(const char *, const char *);
+const char *srs_decode(const char *);
+
+
 /* ssl_smtpd.c */
 void   *ssl_mta_init(void *, char *, off_t, const char *);
 void   *ssl_smtp_init(void *, int);
@@ -1446,29 +1642,30 @@ struct stat_value *stat_timespec(struct timespec *);
 
 
 /* table.c */
-struct table *table_find(const char *, const char *);
-struct table *table_create(const char *, const char *, const char *,
+struct table *table_find(struct smtpd *, const char *);
+struct table *table_create(struct smtpd *, const char *, const char *,
     const char *);
 int	table_config(struct table *);
 int	table_open(struct table *);
 int	table_update(struct table *);
 void	table_close(struct table *);
+void	table_dump(struct table *);
 int	table_check_use(struct table *, uint32_t, uint32_t);
 int	table_check_type(struct table *, uint32_t);
 int	table_check_service(struct table *, uint32_t);
-int	table_lookup(struct table *, struct dict *, const char *, enum table_service,
+int	table_match(struct table *, enum table_service, const char *);
+int	table_lookup(struct table *, enum table_service, const char *,
     union lookup *);
-int	table_fetch(struct table *, struct dict *, enum table_service, union lookup *);
-void table_destroy(struct table *);
+int	table_fetch(struct table *, enum table_service, union lookup *);
+void table_destroy(struct smtpd *, struct table *);
 void table_add(struct table *, const char *, const char *);
 int table_domain_match(const char *, const char *);
 int table_netaddr_match(const char *, const char *);
 int table_mailaddr_match(const char *, const char *);
-void	table_open_all(void);
-void	table_dump_all(void);
-void	table_close_all(void);
-int table_parse_lookup(enum table_service, const char *, const char *,
-    union lookup *);
+int table_regex_match(const char *, const char *);
+void	table_open_all(struct smtpd *);
+void	table_dump_all(struct smtpd *);
+void	table_close_all(struct smtpd *);
 
 
 /* to.c */
@@ -1485,7 +1682,6 @@ const char *sa_to_text(const struct sockaddr *);
 const char *ss_to_text(const struct sockaddr_storage *);
 const char *time_to_text(time_t);
 const char *duration_to_text(time_t);
-const char *relayhost_to_text(const struct relayhost *);
 const char *rule_to_text(struct rule *);
 const char *sockaddr_to_text(struct sockaddr *);
 const char *mailaddr_to_text(const struct mailaddr *);
@@ -1503,12 +1699,13 @@ void addargs(arglist *, char *, ...)
 	__attribute__((format(printf, 2, 3)));
 int bsnprintf(char *, size_t, const char *, ...)
 	__attribute__((format (printf, 3, 4)));
-int mkdirs(char *, mode_t);
 int safe_fclose(FILE *);
 int hostname_match(const char *, const char *);
 int mailaddr_match(const struct mailaddr *, const struct mailaddr *);
 int valid_localpart(const char *);
 int valid_domainpart(const char *);
+int valid_domainname(const char *);
+int valid_smtp_response(const char *);
 int secure_file(int, char *, char *, uid_t, int);
 int  lowercase(char *, const char *, size_t);
 void xlowercase(char *, const char *, size_t);
@@ -1520,20 +1717,29 @@ int rmtree(char *, int);
 int mvpurge(char *, char *);
 int mktmpfile(void);
 const char *parse_smtp_response(char *, size_t, char **, int *);
-void *xmalloc(size_t, const char *);
-void *xcalloc(size_t, size_t, const char *);
-char *xstrdup(const char *, const char *);
-void *xmemdup(const void *, size_t, const char *);
+int xasprintf(char **, const char *, ...)
+    __attribute__((__format__ (printf, 2, 3)));
+void *xmalloc(size_t);
+void *xcalloc(size_t, size_t);
+char *xstrdup(const char *);
+void *xmemdup(const void *, size_t);
 char *strip(char *);
-void iobuf_xinit(struct iobuf *, size_t, size_t, const char *);
-void iobuf_xfqueue(struct iobuf *, const char *, const char *, ...);
+int io_xprint(struct io *, const char *);
+int io_xprintf(struct io *, const char *, ...)
+    __attribute__((__format__ (printf, 2, 3)));
 void log_envelope(const struct envelope *, const char *, const char *,
     const char *);
 int session_socket_error(int);
 int getmailname(char *, size_t);
 int base64_encode(unsigned char const *, size_t, char *, size_t);
 int base64_decode(char const *, unsigned char *, size_t);
+int base64_encode_rfc3548(unsigned char const *, size_t,
+		      char *, size_t);
+void xclosefrom(int);
 
+void log_trace_verbose(int);
+void log_trace(int, const char *, ...)
+    __attribute__((format (printf, 2, 3)));
 
 /* waitq.c */
 int  waitq_wait(void *, void (*)(void *, void *, void *), void *);
@@ -1544,8 +1750,7 @@ void waitq_run(void *, void *);
 struct runq;
 
 int runq_init(struct runq **, void (*)(struct runq *, void *));
-int runq_schedule(struct runq *, time_t, void (*)(struct runq *, void *), void *);
-int runq_delay(struct runq *, unsigned int, void (*)(struct runq *, void *), void *);
-int runq_cancel(struct runq *, void (*)(struct runq *, void *), void *);
-int runq_pending(struct runq *, void (*)(struct runq *, void *), void *, time_t *);
-int runq_next(struct runq *, void (**)(struct runq *, void *), void **, time_t *);
+int runq_schedule(struct runq *, time_t, void *);
+int runq_schedule_at(struct runq *, time_t, void *);
+int runq_cancel(struct runq *, void *);
+int runq_pending(struct runq *, void *, time_t *);
